/*
 *  @(#) PrcFile.java 1.1 25-jun-99
 *
 *  Copyright by <ADRESS>Thomas Holland, innot@frankfurt.netsurf.de</ADRESS>
 *  This is part of the <I>presdec</I> application.
 *  For detailed information read the <TT>README</TT> file supplied with this application.
 *
 */


package presdec;

import java.io.*;
import java.util.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.GregorianCalendar;

import presdec.resource.*;
import presdec.sort.Sorter;
import presdec.sort.ResourceComparer;

/**
 * 	PrcFile
 *
 *	Container for prc files. Contains methods to split the file into separate resources and save
 *	them in numerous ways. 
 *
 *
 *	@author Thomas Holland
 *	@version 1.1  25-jun-99		Added code for dumping resources
 *  	@version 1.0  30-sep-98		initial Version
 *
 *	@see	presdec.Main
 */
public class PrcFile {

// Data members ----------------------------------------------------------

	static PrcFile instance;

// protected data members:

	/** The options as passed from main() */
	protected Hashtable 	options;

	/** shortcut for the option */
	protected boolean	verbose;

	/** Length of the PRC file */
	protected int 		file_length;

	/** Name of the PRC file including the path */
	protected File 		path_name;

	/** The whole PRC file */
	protected byte[] 	prcfile;

	/** Array of all resources */
	protected Vector	p_res;

	// The PRC Header fields
	/** Name of the PRC file as defined in the PRC file header */
	protected String 	hName;

	/** Version of the PRC File (PalmOS version) */
	protected int 		hVersion;

	/**	Date of Creation */
	protected GregorianCalendar hCreatDate;

	/** Creator Type */
	protected String 	hCreator;

	/** Template for Date/Time strings */
	protected SimpleDateFormat dateformat = new SimpleDateFormat("HH:mm:ss dd-MMM-yyyy");

// final static data member:

	/** Size of the PRC file header */
	final static int PRC_HEADER_SIZE = 0x4e;

	/** Size of struct ResourceMapEntry */
	final static int RME_SIZE = 10;

// End of data members ---------------------------------------------------

// Constructor -----------------------------------------------------------

	/**
	 *	The constructor will read the PRC file 
	 *	and split it into Resource objects
	 *
	 *	If any error occurs an appropriate Exception will be thrown.
	 *
	 *	@param options	Hashtable with all options
	 *
	 *	@exception Exception 	Either IOExceptions while reading 
	 *				or ArrayOutOfBoundsException while parsing the prc file.
	 *
	 *	@see	presdec.Main
	 */
	public PrcFile( Hashtable options ) throws Exception {

		instance = this;		// Static reference to ourself

		this.options = options;

		verbose = options.containsKey("verbose");
		
		path_name = (File)options.get("prcfile");

		prc_open();			// open prc file and read everything
		
		test_for_prc(); 		// Do we really have a PRC file?

		divide_into_resources();	// Divide the file into resource objects

    }

// End of constructor ----------------------------------------------------	

// Public methods --------------------------------------------------------	

	/**
	 *	Sort the Resources according to Type, id and relations
	 *
	 */
	public void sort() {
		if(verbose) System.out.println("    sorting resources");

		// First: sort according to order and ID
		try {
			Sorter.sort(p_res, new ResourceComparer());
		} 
		catch (Exception exc) {	// Should not be possible!
			System.out.println("presdec: Internal Error, please contact the author!");
			System.out.println(exc);
			System.exit(-1);
		}
		// Sort according to relations (Help strings and menus behind the assoc. form)
		Resource form;
		Object tmp;
		Enumeration rel; 
		int idx;

		// look for tFRM objects in reverse order, so that multiple used menus or help strings
		// get sorted to the first form using them
		for (int i=p_res.size()-1 ; i>=0; --i) {
			form = (Resource)p_res.elementAt(i);
			if(form.type == "tFRM") {
				if(options.containsKey("sort")) { // sort Form objects to y and x 
					form.sort();
				}
	
				rel = form.getRelated(); // find all related resources
				while(rel.hasMoreElements()) { 
					idx = p_res.indexOf(rel.nextElement());
					if(idx >= 0) {	// Element exists
						tmp = p_res.elementAt(idx);
						p_res.removeElementAt(idx);
						p_res.insertElementAt(tmp, i+1);
					}					
				}
			}
		}
	}

	/**
	 *	Write all resources as binary
	 *
	 *	@exception	IOException	IO Errors
	 *
	 */
	public void saveRes() throws IOException {
		Resource res;

		File path = (File)options.get("path");

		if(verbose) System.out.println("    saving binary resources");

		for(int i=0; i<p_res.size(); i++) {
			res = (Resource)p_res.elementAt(i);
			// save only if either option "binary" true (all resources)
			// or if resource is no rcp item
			if(((Boolean)options.get("binary")).booleanValue() || (!res.rcpitem)) { 
				res.saveRes(path.toString() + path.separator+res.genResFilename());
				if(verbose) System.out.println("        saved " + res.type + " ID " + res.id);
			}
		}
	}


	/**
	 *	Dump resources in a readable way. The actual output depends on the resources,
	 *	e.g. data resources are printed as hexdump and additional reloc 
	 *	information.
	 *
	 *	@exception	IOException	IO Errors
	 *
	 */
	public void dumpRes() throws IOException {
		Resource res;

		File path = (File)options.get("path");

		if(verbose) System.out.println("    dumping resources");

		for(int i=0; i<p_res.size(); i++) {
			res = (Resource)p_res.elementAt(i);
			if(res.dumpitem) {
				res.dumpRes(path.toString() + path.separator + res.genResFilename(".txt"));
				if(verbose) System.out.println("        dumped " + res.type + " ID " + res.id);
			}
		}
	}

	/**
	 *	Generate the rcp file using the supplied writer
	 *
	 *	@param 		writer		RCP file
	 *
	 * 	@exception	Exception 	IO Errors
	 *
	 */
	public void writeRCP(Writer rcpwriter) throws Exception {
 		Resource res;
		Date date = new Date();

		if(verbose) System.out.println("    writing " + options.get("output"));

		rcpwriter.write("// This file was automatically generated from " + path_name.getName()+"\n");
		rcpwriter.write("// by \"presdec\" version 1.2 at " + dateformat.format(date) + "\n");
		rcpwriter.write("// presdec is (c) T.Holland\n");
		rcpwriter.write("// For more information and the latest version contact innot@frankfurt.netsurf.de\n");
		rcpwriter.write("// or visit http://people.frankfurt.netsurf.de/innot/presdec\n");
		rcpwriter.write("\n\n");

		for(int i=0; i<p_res.size(); i++) {
			res = (Resource)p_res.elementAt(i);
			if(res.rcpitem) {
				res.saveRCP(rcpwriter);
				if(verbose) System.out.println("        wrote " + res.type + " ID " + res.id);
			}
		}
	}
 
	/**
	 *	return an instance of this PrcFile object. This is used to access data members
	 *	from outside this object and its descendantes.
	 *	<p>
	 *	Example to get the options from outside of PrcFile
	 *	<code>
	 *		Hashtable options = PrcFile.sharedInstance().getOptions();
	 *	</code>
	 *
	 *	@see #getOptions
	 */
  	public static PrcFile sharedInstance() {
		return instance;
  	}

	/**
	 *	return a reference to the options that were passed to PrcFile from Main.
	 *	The options can be accessed from outside this class by using the 
	 *	<code>sharedInstance()</code> method of this class.
	 *
	 *	@return		a Hashtable containing options
	 *
	 *	@see		presdec.Main
	 *	@see		#sharedInstance
	 */
	public Hashtable getOptions() {
		return options;
	}

	/**
	 *	Return a nicely formatted table of all resources, together with some header information
	 *
	 */
	public String toString() {
		Resource res;
		StringBuffer buf = new StringBuffer();
		
		buf.append("\npresdec analysis of file " + path_name +"\n\n");
		buf.append("----- Header Information -----\n\n");
		buf.append("      Internal name: " + hName + "\n");
		buf.append("            Version: " + hVersion);
		if(hVersion == 1)      buf.append(" (PalmOS V1.0 and V2.0)\n");
		else if(hVersion == 2) buf.append(" (PalmOS V3.0)\n");
		else                   buf.append(" (Unknown PalmOS version)\n");
		buf.append("      Creation Date: " + dateformat.format(hCreatDate.getTime()) + "\n");
		buf.append("       Creator Type: " + hCreator + "\n");
		buf.append("               Size: " + file_length + "\n");
		buf.append("Number of resources: " + p_res.size() + "\n\n");

		buf.append("----- Resources -----\n\n");
		buf.append(" Nr. Type    ID Offset  Size   Data\n");
		buf.append("--------------------------------------------\n");
		
		for (int i=0 ; i<p_res.size(); i++) {
			buf.append(Resource.align(i, 3) + ". ");
			res = (Resource)p_res.elementAt(i);
			buf.append(res.toString() + "\n");
		}
		buf.append("--------------------------------------------\n");

		return buf.toString();
	}

// End of public methods -------------------------------------------------

// Private methods -------------------------------------------------------

	/** 
	 * 	Open the PRC file and read the whole file
	 *
	 *	The filename is stored in <code>path_name</code>
	 *	The data is stored in <code>prcfile</code>
	 *
	 *	@exception	If anything goes wrong
	 *
	 */
	private void prc_open() throws Exception {

		RandomAccessFile raf;

		try {
			raf = new RandomAccessFile( path_name, "r" );
		}
		catch (Exception exc) {
			throw new Exception(path_name + " could not be opened");
		}

		try {
			file_length = (int) raf.length();
			prcfile = new byte[file_length];

			raf.seek( 0 );
			int bytes_read = raf.read( prcfile );
			if ( bytes_read < file_length )
				throw new Exception(path_name + " could not be read complete ("+(file_length-bytes_read)+" bytes short)");
		}
		catch ( Exception exc ) {
			throw new Exception(path_name + " could not be read");
		}
	}


	/**
	 *	Divide the PRC bytearray into resource objects
	 *
	 *	@exception 		Any error in generating the resources
	 *
	 */
	private void divide_into_resources() throws Exception {
		// get the number of resources in the file,
    		// stored in the last ulong of the header
		int c_res = int_4byte(prcfile, PRC_HEADER_SIZE - 4);

		// a little sanity check
		if ( c_res <0 || c_res > ((file_length-PRC_HEADER_SIZE)/RME_SIZE)) {
			throw new Exception("Error in PRC file: impossible number of resources");
		}

		p_res = new Vector(c_res);

		// Common for all resources
		StringBuffer type = new StringBuffer("aaaa");
		int id = 0;
		int offset =0;
		int size = 0;

		for(int i=0; i<c_res; i++) {
			// p is a pointer to the current resource
			int p = PRC_HEADER_SIZE + i * RME_SIZE;

			// Get the type, id and offset
			for(int n=0 ;n<4; n++) type.setCharAt(n, (char)prcfile[p+n]);
			id = int_2byte(prcfile, p+4);
			offset = int_4byte(prcfile, p+6);
			
			if (offset < PRC_HEADER_SIZE || offset > file_length) {
				throw new Exception("Error in PRC file: incorrect offset in RME at " + p);
			}
 
	    		// size is a little bit more difficult, as it is not stored in the 
	    		// ResourceMapEntry. Size is Offset of the next RME - the current Offset.
	    		// For the last RME we take the filesize.

			if (i < (c_res - 1))  {	// not the last RME
				size = int_4byte(prcfile, p+6+RME_SIZE) - offset;	
			}
			else  {
				size = file_length - offset;
			}
	    
			if (size > (file_length-PRC_HEADER_SIZE)) {
				throw new Exception("Error in PRC file: incorrect size in RME at " + p);
			}
 	
			// We got all we need, so lets create the appropiate resource object
			if(type.toString().equals("MBAR")) { // Pulldown Menu
				p_res.addElement(new Res_MBAR(id, prcfile, offset, size));
			}
			else if(type.toString().equals("APPL")) { // Application 
				p_res.addElement(new Res_APPL(id, prcfile, offset, size));
			}
			else if(type.toString().equals("tAIN")) { // Application Icon Name
				p_res.addElement(new Res_tAIN(id, prcfile, offset, size));
			}
			else if(type.toString().equals("tAIB")) { // Application Icon Bitmap
				p_res.addElement(new Res_tAIB(id, prcfile, offset, size));
			}
			else if(type.toString().equals("tFRM")) { // Form
				p_res.addElement(new Res_tFRM(id, prcfile, offset, size));
			}
			else if(type.toString().equals("tver")) { // Version string
				p_res.addElement(new Res_tver(id, prcfile, offset, size));
			}
			else if(type.toString().equals("tSTR")) { // Generic String
				p_res.addElement(new Res_tSTR(id, prcfile, offset, size));
			}
			else if(type.toString().equals("Talt")) { // Alert
				p_res.addElement(new Res_Talt(id, prcfile, offset, size));
			}
			else if(type.toString().equals("Tbmp")) { // Bitmap
				p_res.addElement(new Res_Tbmp(id, prcfile, offset, size));
			}
			else if(type.toString().equals("data")) { // Data segment
			        p_res.addElement(new Res_data(id, prcfile, offset, size));
			}
			else { // Default for unknown resources
				p_res.addElement(new Resource(type.toString(), id, prcfile, offset, size));
			}
		} // for loop
	} // divide_into_resources()


	/** 
	 *	Checks for a correct PRC file header (and gets some information from the header)
	 *
	 *	It will do so by checking that the type contains 'appl' and the flag field contains 0x01. 
	 *	If all are correct <code>test_for_prc()</code>
	 *	returns, otherwise it throws an Exception
	 *
	 *	As a side effect <code>test_for_prc</code> will fetch some useful information from the header,
	 *	name, version, creation time and creator
	 *
	 * 	@see 	presdec.PrcFile#hName
	 *	@see	presdec.PrcFile#hVersion
	 *	@see	presdec.PrcFile#hCreatTime
	 *	@see	presdec.PrcFile#hCreator
	 *
	 *	@exception	If PRC file header is not correct
	 *
	 */ 
	private void test_for_prc() throws Exception {
		// We check that flag field is 0x01 and that the type is 'appl'
		if(int_2byte(prcfile, 32) == 0x0001) {
			if(int_4byte(prcfile, 60) == 0x6170706c) {
				// The check is complete, lets get the other fields
				StringBuffer tmp = new StringBuffer(32);

				// get Name...
				for(int i=0; prcfile[i]!=0; i++) {
					tmp.append((char)prcfile[i]);
				}
				hName = tmp.toString();

				// ...Version...
				hVersion = int_2byte(prcfile, 34);

				// ...creation date...
				hCreatDate = new GregorianCalendar(1904,0,1); // Mac Epoch starts midnight, jan 1st 1904
				int seconds = int_4byte(prcfile, 36); // The number of seconds will usually be >0x8000 and therefore negative
				if(seconds<0) {	// split the add in two parts
					hCreatDate.add(Calendar.SECOND, 0x7fffffff);
					seconds = seconds - 0x7fffffff;
				}
				hCreatDate.add(Calendar.SECOND, (int)seconds);

 				// ...and Creator.
				tmp = new StringBuffer(4);
				tmp.append((char)prcfile[64]);
				tmp.append((char)prcfile[65]);
				tmp.append((char)prcfile[66]);
				tmp.append((char)prcfile[67]);
				hCreator = tmp.toString();

				return;
			}
		}
		throw new Exception(path_name + " has a incorrect header");
	}

	/** 
	 *	Extract an long (4 byte) integer out of a byte array 
	 *
	 *	Note: No error checking
	 *
	 *	@params	b	The byte array 
	 *	@params start	Offset of the 4 byte integer within the array
	 *
	 *	@returns	The generated integer
	 *
	 */
	private int int_4byte ( byte[] b, int start ) {
		int result;
		result =  ((int)b[start+3])&0x000000FF;
		result +=(((int)b[start+2])&0x000000FF)<<8;
		result +=(((int)b[start+1])&0x000000FF)<<16;
		result +=(((int)b[start])  &0x000000FF)<<24;
		return result;
	}

	/** 
	 *	Extract an short (2 byte) integer out of a byte array 
	 *
	 *	Note: No error checking
	 *
	 *	@params	b		The byte array 
	 *	@params start	Offset of the 2 byte integer within the array
	 *
	 *	@returns		The generated integer
	 *
	 */
	private int int_2byte ( byte[] b, int start ) {
		int result;
		result =  ((int)b[start+1])&0x000000FF;
		result +=(((int)b[start])  &0x000000FF)<<8;
		return result;
	}

} // end of PrcFile class
