/*
 *  @(#) ResByte.java 1.0 30-sep-98
 *
 *  Copyright by <ADRESS>Thomas Holland, innot@frankfurt.netsurf.de</ADRESS>
 *  This is part of the <I>presdec</I> application.
 *  For detailed information read the <TT>README</TT> file supplied with this application.
 *
 */

package presdec.resource;

/**
 *  ResByte Class
 *  <p>
 *  A kind of ByteArray with specialized methods for parsing PalmPilot resources
 *
 *  @author Thomas Holland
 *  @version 1.0  30-sep-98
 *
 *  @see	presdec.resource.Resource
 */
public class ResByte {

// Data members ----------------------------------------------------------

	/** Binary image of the resource */
	protected byte		data[];

	/** Size of the array */
	protected int 		size;

	/** Pointer used for stepping through the resource */
	protected int 		ptr;

// End of data members ---------------------------------------------------
// Constructors ----------------------------------------------------------

	/** 
	 *  Construct a ResByte object out of a byte array
	 *
	 *  @param in		Bytearray to be copied
	 *
	 *  @return		void
	 *
	 */
	ResByte(byte[] in) {
		this(in, in.length, 0);
	}

	/** 
	 *  Construct a ResByte object out of a byte array, starting at an offset
	 *
	 *  @param in		Bytearray to be copied
	 *  @param size		the size of the new array
	 *  @param offset	start copying at this offset
	 *
	 *  @return		void
	 *  
	 */
	ResByte(byte[] in, int size, int offset) {
		data = new byte[size];
		System.arraycopy(in, offset, data, 0, size);
		ptr = 0;		// set pointer to beginning of the byte array
		this.size = size; 
	}

// End of Constructors ---------------------------------------------------

	/**
	 *  Set the data pointer to some new value
	 *
	 *  @param offset	New offset within byte array
	 *
	 *  @returns		the new value
	 *
	 *  @exceptions		ArrayIndexOutOfBoundsException
	 *	
	 */
	public int setPtr(int offset) throws ArrayIndexOutOfBoundsException {
		if(offset > size ) throw new ArrayIndexOutOfBoundsException("setPtr(" + offset +"): invalid index");
		return ptr = offset;
	}

	/** 
	 *  Return <code>data[]</code> as array of bytes
	 *
	 *  @returns		The data array
	 *
	 */
	public byte[] toByteArray() {
		return data;
	}

	/**
	 *  Get the current vaue of the data pointer
	 *
	 *  @returns		The current value of <code>ptr</code>
	 *
	 */
	public int getPtr() {
		return ptr;
	}

	/**
	 *  Fetch a byte from the data
	 *
	 *  @returns		Next Byte (as int)
	 *
	 */
	public int getNextByte() {
		return ((int)data[ptr++])&0x000000FF;
	}

	/**
	 *  Fetch a word from the data
	 *
	 *  @returns		Next Word (as int)
	 *
	 */
	public int getNextWord() {
		int result;
		result = (((int)data[ptr++])&0x000000FF)<<8;
		result += ((int)data[ptr++])&0x000000FF;
		return result;
	}

	/**
	 *  Fetch a long from the data
	 *
	 *  @returns		Next Long (as int)
	 *
	 */
	public int getNextLong() {
		int result;
		result = (((int)data[ptr++])&0x000000FF)<<24;
		result +=(((int)data[ptr++])&0x000000FF)<<16;
		result +=(((int)data[ptr++])&0x000000FF)<<8;
		result += ((int)data[ptr++])&0x000000FF;
		return result;
	}
		
	/**
	 *  Fetch a string from the data
	 *
	 *  PalmPilot Resources use C type strings, which are converted to String objects
	 *
	 *  @param	conv	Convert unprintable characters to string literals 
	 *
	 *  @returns		Next string in the data
	 *
	 */
	public String getNextString(boolean conv) {
		String str;
		int i;

		// if we dont need to convert then use the quick method
		if(!conv) {
			// find the next "\0" in the data
			for(i=0; data[ptr+i] != 0; i++);
			str = new String(data, ptr, i);
			ptr += i+1;	// step over the "\0" at the end of the string
			return str;
		}
		else {
			StringBuffer buf = new StringBuffer();
			int c;
			for(i=0; (c = (int)(data[ptr+i] & 0xff)) != 0; i++) {
				if((c<32 || c>127) && c!=10) {	// something unprintable
					// kludge to make all octal strings 3 digits long
					if(c<8) buf.append("\\00");
					else if(c<64) buf.append("\\0");
					else buf.append("\\");
					buf.append(Integer.toOctalString(c));
				}
				else {
					if(c == 34 || c == 92) {
						// Escape backslashes and quotation marks
						buf.append("\\");						}
					buf.append((char)c);
				}
			}
			ptr += i+1;
			return buf.toString();
		}	
	}

	/**
	 *  Fetch a string from the data at a certain offset
	 *
	 *  PalmPilot Resources use C type strings, which are converted to String objects
	 *
	 *  @param	conv		Convert unprintable characters to string literals
	 *  @param	offset		start of the string within the byte array
	 *
	 *  @returns			Next string in the data
	 *
	 */
	public String getString(int offset, boolean conv) {
		String str;
		int i;

		// if we dont need to convert then use the quick method
		if(!conv) {
			// find the next "\0" in the data
			for(i=0; data[offset+i] != 0; i++);
			str = new String(data, offset, i);
			return str;
		}
		else {
			StringBuffer buf = new StringBuffer();
			int c;
			for(i=0; (c = (int)(data[offset+i] & 0xff)) != 0; i++) {
				if((c<32 || c>127) && c!=10) {	// something unprintable
					// kludge to make all octal strings 3 digits long
					if(c<8) buf.append("\\00");
					else if(c<64) buf.append("\\0");
					else buf.append("\\");
					buf.append(Integer.toOctalString(c));
				}
				else {
					if(c == 34 || c == 92) {
						// Escape backslashes and quotation marks
						buf.append("\\");						}
					buf.append((char)c);
				}
			}
			return buf.toString();
		}	
	}

	/**
	 *  Divides a String into seperate lines, not longer than <code>max</code>
	 *  chars long, paying special attention to newlines
	 *
	 *  @param str		String to be divided
	 *  @param max		Maximum length of line, excluding the indent
	 *  @param indent	Amount of indent
	 *
	 *  @returns		The new String, suitable for a Writer
	 *
	 */
	public static String splitString(String str, int max, int indent) {
	
		StringBuffer sbuf = new StringBuffer();
		int cursor = 0;
		int len	= str.length();

		do {
			// first look for a newline or a 0x00 string separator
			int nl = str.indexOf(10, cursor);
			int ss = str.indexOf(0, cursor);
			if (ss != -1) { // Take 0x00 seperator if it comes first
				if(ss<nl || nl==-1) {
					nl = ss;
				}
			}
			if (nl == -1 || nl-cursor>max) { // no seperator in this line or none at all
				if( len-cursor <= max ) { // we dont need to split at all
					sbuf.append("\"");
					sbuf.append(str.substring(cursor, len));
					sbuf.append("\"\n");
					cursor = len;
				}
				else {	// split at the last space before max chars
					int ls = str.lastIndexOf(32, (cursor+max)>len?len:cursor+max);
					sbuf.append("\"");
					sbuf.append(str.substring(cursor, ls+1));
					sbuf.append("\"\\\n");
					for(int i=0; i<indent; i++)
						sbuf.append(" ");
					cursor = ls + 1;
				}
			}
			else {	// We need to split at a seperator
				String sep = null;
				sbuf.append("\"");
				sbuf.append(str.substring(cursor, nl));
				switch((int)str.charAt(nl)) {	// get the correct seperator 
					case 10:	sep = "\\n";
								break;
					case 0:		sep = "\\000";
								break;
				}
				if(nl < len-1) { // append a '\' only if not at the end of the string
					sbuf.append(sep + "\"\\\n");
					for(int i=0; i<indent; i++)
						sbuf.append(" ");
				}
				else
					sbuf.append(sep + "\"\n");
				cursor = nl + 1;

			}
		} while(cursor<len);

		return sbuf.toString();
	}

	/**
	 *  Generate a String with size and ptr 
	 */
	public String toString() {
		return "ResByte: Size " + size + "  ptr " + ptr;
	}

// End of public methods -------------------------------------------------

} // end of class ResByte


