/*
 *  @(#) Resource.java 1.1 25-jun-98
 *
 *  Copyright by <ADRESS>Thomas Holland, innot@frankfurt.netsurf.de</ADRESS>
 *  This is part of the <I>presdec</I> application.
 *  For detailed information read the <TT>README</TT> file supplied with this application.
 *
 */

package presdec.resource;
 
import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import presdec.PrcFile;
import presdec.resource.ResByte;


/**
 *  Resource Class
 *
 *  Parent class for all resources contained in a PRC file.
 *  Containes the common data members for all resources and
 *  code to save a resource as binary as well as some support
 *  methods needed by some subclasses.
 *
 *  @author Thomas Holland
 *  @version 1.1  25-jun-98 Added the dumpRes method to dump the resource in human readable form
 *
 *
 */
public class Resource {

// Data members ----------------------------------------------------------

	/** Binary image of the resource */
	protected ResByte	resource;

	/** Type of the resource */
	public String 		type;

	/** ID of the resource */
	public int     		id;

	/** Size of the binary image */
	protected int 		size;

	/** Location of the resource in the PRC file */
	protected int		offset;

	/** Set true if the resource can be saved in a rcp file */
	public boolean		rcpitem;

	/** Set true if the resource can be dumped */
	public boolean		dumpitem;
	
	/** Flag for faulty resources */
	public boolean		resOK;

	/** The order to sort the resources */
	public int 		order;

	final static int 	APPLICATION = 1;
	final static int 	APPLICATIONICONNAME = 2;
	final static int	APPLICATIONICON = 3;
	final static int 	VERSION = 5;
	final static int 	FORM = 10;
	final static int 	MENU = 10;
	final static int 	STRING = 10;
	final static int 	ALERT = 10;
	final static int 	BITMAP = 12;
	final static int 	TRAP = 20;
	final static int 	CODE = 90;
	final static int	DATA = 95;
	final static int	OTHER = 100;
	final static int 	VIRTUAL = 999;	// if this object is only used for comparing

// End of data members ---------------------------------------------------
// Constructors ----------------------------------------------------------

	/** 
	 *	Construct a Resource object out of a byte array
	 *
	 *	Since some data is stored in the ResourceMapEntry and not in the 
	 *	Resource itself they have to be passed as arguments to the constructor
	 *
	 *	@param type	Type of resource as 4 letter String
	 *	@param id	ID of the resource
	 *	@param prc	Bytearray of the PRC file
	 *  	@param offset   Starting point of the resource in the prc file
	 *  	@param size	Size of the resource
	 *
	 *  	@return		void
	 *  
	 */
	public Resource(String type, int id, byte[] prc, int offset, int size) {
		// make local copies of the arguments
		this.type 		= type;
		this.id 		= id;
		this.size 		= size;
		this.offset 		= offset;

		// should be overridden in children classes
		this.rcpitem 		= false;
		this.dumpitem		= true;
		this.order	 	= OTHER;

		// Allocate the resource and copy from the prc file
		resource = new ResByte(prc, size, offset);

		resOK = true; 	// and assume that the resource is OK
	}

	/**
	 *	Short Constructor to build something that can be used for the equalTo method
	 *	
	 *	@param type		Type of resource as 4 letter String
	 *	@param id		ID of the resource
	 *
	 *  	@return			void
	 *  
	 */
	public Resource(String type, int id) {
		this.type 	= type;
		this.id 	= id;
		this.order 	= VIRTUAL;
	}
 

// End of Constructors ---------------------------------------------------
// Public methods --------------------------------------------------------

	/**
	 *	Dump the resource in some human readable way. This method needs to be
	 *	overwritten by children that actually have something to dump.
	 *
	 *  	@param filename	filename
	 *
	 *	@return		boolean success
	 * 
	 *	@exception	IOException If the resource could not be written 
	 *
	 */
	public boolean dumpRes(String filename) throws IOException {
		
		return false;
	}
	
	
	/**
	 *	Save Resource as binary, using standard name.
	 *
	 *	@return		boolean success
	 * 
	 *	@exception	IOException If the resource could not be written 
	 *
	 */
	public boolean saveRes() throws IOException {
		return saveRes(genResFilename());
	}

	/**
	 *	Save Resource as binary, using the supplied filename.
	 *
	 *  	@param filename	optional filename
	 *
	 *	@return		boolean success
	 * 
	 *	@exception	IOException If the resource could not be written 
	 *
	 */
	public boolean saveRes(String filename) throws IOException {
		FileOutputStream resout = new FileOutputStream(filename);

		resout.write(resource.toByteArray());
		return true;
	}

	/** 
	 * 	Generate a Filename with ".bin" extension for this resource
	 *
	 *	@return 	Filename in the form: ttttxxxx 
	 *			with type t and ID x as hex
	 *
	 */
	public String genResFilename() {
 		return genResFilename(".bin");
	}


	/** 
	 * 	Generate a Filename with the specified extension for this resource
	 *
	 *	@param		The extension to be added to the filename. Can be empty
	 *
	 *	@return 	Filename in the form: ttttxxxx 
	 *			with type t and ID x as hex
	 *
	 */
	public String genResFilename(String extension) {
		StringBuffer hex = new StringBuffer();
		hex.append(Integer.toHexString((id & 0xf000)>>12)); 
		hex.append(Integer.toHexString((id & 0x0f00)>>8));
		hex.append(Integer.toHexString((id & 0x00f0)>>4)); 
		hex.append(Integer.toHexString((id & 0x000f))); 
 		return type + hex.toString() + extension;
	}

	/**
	 *	Save Resource as RCP elements
	 * 
	 *	Every Resource that can be saved as ascii should override this method.
	 * 
	 *
	 *	@param writer	where the ascii description should be saved
	 *	
	 * 	@return 	boolean if the resource is not saveable in a RCP
	 *					
	 * 	@exception	java.io.IOException If the resource could not be written
	 *
	 */
	public boolean saveRCP(Writer writer) throws java.io.IOException {
		return false;
	}

	/**
	 *	Sort Resource objects (only necessary for tFRM resources)
	 *
	 */
	public void sort() {
	}

	/** 
	 *	Get a list of resources that should follow this resource in a sorted RCP file
	 *
	 *	This is only used by the tFRM resource
	 *
	 *	@return		Vector of resources that should follow
	 *
	 */
	public Enumeration getRelated() {
		Vector v = new Vector();
		return v.elements();
	} 

	/**
	 *	Compare this Resource with another resource
	 *	
	 *	@param 	anObject	Object to compare to
	 *
	 *	@return			true if both objects equal in type and id
	 *
	 */
	public boolean equals(Object anObject) {
		if((anObject != null) && (anObject instanceof Resource)) {
			Resource res = (Resource)anObject;
			if(this.id == res.id) {
				if(this.type.equals(res.type)) {
					return true;
				}
			}
		}
		return false;
	}


	/**
	 *	Generate a String with Type, ID and Size of the resource for debugging
	 */
	public String toString() {
		StringBuffer buf = new StringBuffer(70);
				
		buf.append(type + align(id,6) + align(offset,7) + align(size,6));

		return buf.toString();
	}

// End of public methods -------------------------------------------------
// Private methods -------------------------------------------------------

	public static String align(int val, int size) {
		
		String value = Integer.toString(val);
		StringBuffer buf = new StringBuffer(value);
		for(int i=value.length(); i<size; i++) buf.insert(0,' ');

		return buf.toString();
	}

// End of private methods ------------------------------------------------

} // end of class Resource

