/*
 *  @(#) BmpConverter.java  1.0 19-jul-99
 *
 *  Copyright by <ADRESS>Thomas Holland, innot@frankfurt.netsurf.de</ADRESS>
 *  This is part of the <I>presdec</I> application.
 *  For detailed information read the <TT>README</TT> file supplied with this application.
 *
 */

package presdec.resource.bitmap;

import java.io.IOException;
import java.io.OutputStream;
import presdec.resource.bitmap.Converter;
import presdec.resource.bitmap.BmpConstants;

/**
 *	Write the PalmOS bitmap resource as a Windows Bitmap file (bmp)
 *  
 *	@author 	Thomas Holland
 *	@version	1.0 19-jul-99
 *	@see 		presdec.resource.Res_Tbmp
 *	@see 		presdec.resource.bitmap.PilotBitmap
 *	@see		presdec.resource.bitmap.Converter
 */

public class BmpConverter implements Converter, BmpConstants {

	/**
	 *	Convert the supplied array of pixels into bmp format and write it out to
	 *	the supplied <code>OutputStream</code>.
	 *	The Bitmap is monochrome (2 colors) and is not compressed to keep this Class simple 
	 *	
	 *	@param		pix	two-dimensional array of pixel values (for b/w bitmaps a pixel
	 *				may be <code>0</code> or <code>1</code>.
	 *	@param		x	number of columns in the pixel array.
	 *	@param		y	number of rows in the pixel array.
	 *	@param		depth	number of bits per pixel, currently set to 1 (b/w).
	 *	@param		outs	to write the generated data into.
	 *
	 *	@exception	IOException	as generated by the <code>OutputStream</code>
	 */
	public void save(int pix[][], int x, int y, int depth, OutputStream outs) throws IOException {

		// Build the bitmap file header
		byte[] bmpfileheader = new byte[BMP_HEADERSIZE];
		byte[] bmpinfoheader = new byte[BMP_INFOHEADERSIZE];
		byte[] bmpcolormap   = new byte[BMP_COLORMAPSIZE];
		
		// Calculate the numerous sizes used in a bitmap file
		int bytesperrow = 4+(((x-1)/32)*4);	// Aligned to multiple of 4, with one long for every 32 pixels
		int bitmapsize   = y * bytesperrow;
		int headersize   = BMP_HEADERSIZE + BMP_INFOHEADERSIZE + BMP_COLORMAPSIZE;
		int filesize     = bitmapsize + headersize;
		
		// fill the bitmap file header
		bmpfileheader[BMP_bfType] = 0x42;			// 'BM'
		bmpfileheader[BMP_bfType +1 ] = 0x4d;
		writeLEDWord(filesize, bmpfileheader, BMP_bfSize);
		writeLEDWord(0, bmpfileheader, BMP_bfReserved);
		writeLEDWord(headersize, bmpfileheader, BMP_bfOffBits);	// Bitmap data starts right behind the headers
		
		// fill the bitmap info header
		writeLEDWord(BMP_INFOHEADERSIZE, bmpinfoheader, BMP_biSize);
		writeLEDWord(x, bmpinfoheader, BMP_biWidth);
		writeLEDWord(y, bmpinfoheader, BMP_biHeight);
		writeLEWord(1, bmpinfoheader, BMP_biPlanes);
		writeLEWord(1, bmpinfoheader, BMP_biBitCount);		// monochrome
		writeLEDWord(0, bmpinfoheader, BMP_biCompression);	// not compressed
		writeLEDWord(bitmapsize, bmpinfoheader, BMP_biSizeImage);
		writeLEDWord(0, bmpinfoheader, BMP_biXPelsPerMeter);
		writeLEDWord(0, bmpinfoheader, BMP_biYPelsPerMeter);
		writeLEDWord(0, bmpinfoheader, BMP_biClrUsed);
		writeLEDWord(0, bmpinfoheader, BMP_biClrImportant);
		
		// fill the colormap
		writeColor(bmpcolormap, 0, 0x00, 0x00, 0x00);	// black
		writeColor(bmpcolormap, 1, 0xff, 0xff, 0xff);	// white
		
		// build the bitmap
		byte bmpdata[] = new byte[bitmapsize];
		int  packedpixels;
		int  cnt = 0;
		
		for(int row = y-1 ; row >= 0 ; row--) {
			for(int col = 0 ; col < bytesperrow ; col++) {
				packedpixels = 0;
				for(int i = 7 ; i>=0 ; i--) {
					if((col*8 + (7-i)) < x) {
						if(pix[col*8+(7-i)][row] == 0) 
							packedpixels |= 1<<i;
						else
							packedpixels &= ~(1<<i);
					}
				}
					
				bmpdata[cnt++] = (byte)packedpixels;
			}
		}
		
		// write the whole thing to the output stream
		outs.write(bmpfileheader);
		outs.write(bmpinfoheader);
		outs.write(bmpcolormap);
		outs.write(bmpdata);
	}


	/**
	 *	return file extension for a bitmap file.
	 *	
	 *	@return		String "bmp"
	 *
	 */
	public String getExt() {
		return "bmp";
	}

	/**
	 *	Write a Word (16 bit) in little endian format to the array at the 
	 *	specified offset
	 *	
	 *	@param	data	The 16 bit value to be written
	 *	@param	array	The array into which the word is written
	 *	@param  offset	Where to write the data in the array
	 */
	private void writeLEWord(int data, byte array[], int offset) {
		
		// No error checking done!
		array[offset]   = (byte)((data & 0x000000ff));
		array[offset+1] = (byte)((data & 0x0000ff00)>>8);
	}


	/**
	 *	Write a DWord (32 bit) in little endian format to the array at the 
	 *	specified offset
	 *	
	 *	@param	data	The 32 bit value to be written
	 *	@param	array	The array into which the word is written
	 *	@param  offset	Where to write the data in the array
	 */
	private void writeLEDWord(int data, byte array[], int offset) {
		
		// No error checking done!
		array[offset]   = (byte)((data & 0x000000ff));
		array[offset+1] = (byte)((data & 0x0000ff00)>>8);
		array[offset+2] = (byte)((data & 0x00ff0000)>>16);
		array[offset+3] = (byte)((data & 0xff000000)>>24);
	}


	/**
	 *	Write the RGB values into the colormap at the specified index
	 *	
	 *	@param	colormap	The colormap
	 *	@param	index		Index into the colormap
	 *	@param  red		The red value
	 *	@param  green		The green value
	 *	@param  blue		The blue value
	 */
	private void writeColor(byte colormap[], int index, int red, int green, int blue) {
		
		// No error checking done!
		int offset = index * 4;
		
		colormap[offset++] = (byte)red;
		colormap[offset++] = (byte)green;
		colormap[offset++] = (byte)blue;
		colormap[offset++] = (byte)0x00;
	}

}