/*
 *  @(#) PilotBitmap.java 1.1 19-jun-99
 *
 *  Copyright by <ADRESS>Thomas Holland, innot@frankfurt.netsurf.de</ADRESS>
 *  This is part of the <I>presdec</I> application.
 *  For detailed information read the <TT>README</TT> file supplied with this application.
 *
 */

package presdec.resource.bitmap;
 
import java.io.IOException;
import java.io.OutputStream;
import java.util.Hashtable;
import presdec.resource.ResByte;
import presdec.resource.bitmap.Converter;

/**
 * 	PilotBitmap Resource
 *	<p>
 *	This Class takes a PalmOS Bitmap and converts it into an array of pixels,
 *	unpacking it if necessary. The method <code>saveBitmap</code> can then be called
 *	with a suitable <code>Converter</code> Class to save the array in any kind of bitmap file.
 *	<p>
 *	It is used by the <code>Res_Tbmp</code> and <code>Res_tAIB</code> Classes to store a
 *	PalmOS bitmap.
 *
 *	@author Thomas Holland
 *	@version	1.1 19-jun-99	changed to OutputStream iso Writer for saveing
 *	@version 	1.0  30-sep-98
 *	@see	presdec.resource.bitmap.Converter
 *	@see	presdec.resource.bitmap.PbmAsciiConverter
 *	@see	presdec.resource.Res_Tbmp
 *	@see	presdec.resource.Res_tAIB
 */

public class PilotBitmap {

// Data members ----------------------------------------------------------

	/**
	 *	To make conversion easier we save the bitmap as an array of int pixels.
	 *	Currently only B/W (1/0) supported, but in the future other values might be possible
	 */
	int pix[][];

	/** Size of the bitmap */
	int x, y;

	/** true if the bitmap was compressed */
	public boolean compressed;

	/** Number of bytes in each row of the bitmap, used for compressed bitmaps */
	int bytes_per_row;


// End of data members ---------------------------------------------------
// Constructors ----------------------------------------------------------

	/** 
	 *	Construct a PilotBitmap Object out of a Bitmap resource
	 *
	 *	@param bitmap	The bitmap resource
	 *
	 *  	@return		void
	 *  
	 */
    public PilotBitmap(ResByte bitmap) {

		int tmp;

		x = bitmap.getNextWord();
		y = bitmap.getNextWord();

		bytes_per_row = bitmap.getNextWord();

		tmp = bitmap.getNextWord();
		if ((tmp & 0x8000)==0)
			compressed = false;
		else
			compressed = true;

		// Next 4 Words are unused, set Ptr to beginning of bitmap data
		bitmap.setPtr(16);

		// Room for the bitmap data
		byte data[];

		if(compressed) {	// Unpack data first if necessary
			// Room for the unpacked data
			data = new byte[16+y*bytes_per_row];

			// to make handling easier we convert the ResByte to a array of bytes
			byte comp[] = bitmap.toByteArray();

			// the first word of the data is the size of the compressed data
			int bmpsize = bitmap.getNextWord();
			

			// Each row has the following structure:  flag [bytes] [flag] [bytes] ...
			// Each flag represents 8 Bytes of bitmap data.
			// If a bit in the flag is 1, then one byte of data will follow,
			// if a bit in the flag is 0, then the bitmap data from the previous row is copied.
			
			byte flag = 0;
			int  counter = 18;	// start of the real data
			
			for(int row = 0; row<y; row++) { // start with a loop of rows
				for(int i=0; i<bytes_per_row; i++) { // all bytes in this row
					if((i%8)==0) { // For every eigth bytes is one flag
						flag = comp[counter++];
					}
					if(((flag<<(i%8)) & 0x80)==0) {	// bit is clear, so copy from previous row
						data[row*bytes_per_row + i + 16] = data[(row-1)*bytes_per_row + i + 16];
					}
					else {	// otherwise take from the compressed data
						data[row*bytes_per_row + i + 16] = comp[counter++];
					} 
				}
			}
		}
		else {	// The bitmap is not compressed, so just copy
			data = bitmap.toByteArray();
		}


		// Now we convert the bitmap data into an array of pixels for easier converting
		pix = new int[x][y];
		byte bitdata;

		for(int row = 0; row<y; row++) {
			for(int col = 0; col<x; col++) {
				bitdata = data[16 + (row*bytes_per_row) + (col/8)];
				pix[col][row] = ((bitdata & (0x01<<(7-(col%8))))==0)?0:1; // ? Formula too Complex Error :-)
			}
		}
	}

// End of Constructors ---------------------------------------------------
// Public methods --------------------------------------------------------
	
	/**
	 *	Save the bitmap data using the supplied <code>Converter</code> to a <code>OutputStream</code>.
	 *	<p>
	 *	This method is called by the <code>saveRCP</code> methods from the underlying Classes
	 *	<code>Res_Tbmp</code> and <code>Res_tAIB</code>.
	 *
	 *	@param	outs		where to save the bitmap
	 *	@param	conv		converts the pixel data into a different format
	 *
	 *	@see	presdec.resource.bitmap.Converter
	 */

	public void saveBitmap(OutputStream outs, Converter conv) throws IOException {
		// depth is fixed to 1 (b/w) for now
		conv.save(pix, x, y, 1, outs);
	}

// End of public methods -------------------------------------------------


} // end of class PilotBitmap



